from typing import List
import os
import requests
import json
import asyncio
import aiohttp
import time
import tqdm

sem = asyncio.Semaphore(20)


def query_x_ray_entries(maximum_resolution: float = 1.5) -> List[str]:

    # RCSB search API
    # https://search.rcsb.org/redoc/index.html
    # https://search.rcsb.org/index.html#search-api

    pdb_ids = []

    url = "https://search.rcsb.org/rcsbsearch/v1/query"

    query = {
        "query": {
            "type":
            "group",
            "logical_operator":
            "and",
            "nodes": [{
                "type": "terminal",
                "service": "text"
            }, {
                "type":
                "group",
                "logical_operator":
                "and",
                "nodes": [
                    {
                        "type": "terminal",
                        "service": "text",
                        "parameters": {
                            "attribute": "exptl.method",
                            "operator": "exact_match",
                            "value": "X-RAY DIFFRACTION"
                        }
                    },
                    {
                        "type": "terminal",
                        "service": "text",
                        "parameters": {
                            "operator": "less_or_equal",
                            "value": maximum_resolution,
                            "attribute": "rcsb_entry_info.resolution_combined"
                        }
                    },
                ]
            }]
        },
        "return_type": "entry",
        "request_options": {
            "return_all_hits": True
        }
    }

    response = requests.post(url, data=json.dumps(query))
    response.raise_for_status()

    response_status = response.status_code

    text_content = response.text
    json_data = json.loads(text_content)

    for result in json_data["result_set"]:
        pdb_ids.append(result["identifier"])

    return pdb_ids


async def async_download_pdbx(pdb_id: str, download_dir: str,
                              session: aiohttp.ClientSession,
                              sem: asyncio.Semaphore) -> None:

    pdb_url = "https://files.rcsb.org/download/{}.cif".format(pdb_id)
    pdb_file_path = os.path.join(download_dir, "{}.cif".format(pdb_id))

    # This timeout is useless.
    # timeout = aiohttp.ClientTimeout(total=30)
    # A workaround solution:
    # https://stackoverflow.com/a/64686124
    # https://github.com/aio-libs/aiohttp/issues/3203
    timeout = aiohttp.ClientTimeout(total=None, sock_connect=30, sock_read=30)

    async with sem:
        async with session.get(pdb_url, timeout=timeout) as response:
            if response.status == 200:
                content = await response.read()
                with open(pdb_file_path, "wb") as fhand:
                    fhand.write(content)
                # print("PDBx {} downloaded.".format(pdb_id))
            else:
                print("Unable to download PDBx {}!".format(pdb_id))


async def async_download_pdbxs(pdb_ids: str,
                               download_dir: str,
                               maximum_num_connections: int = 5) -> None:

    if not os.path.exists(download_dir):
        os.mkdir(download_dir)

    tasks = []
    # https://pawelmhm.github.io/asyncio/python/aiohttp/2016/04/22/asyncio-aiohttp.html
    sem = asyncio.Semaphore(5)

    connector = aiohttp.TCPConnector(limit_per_host=maximum_num_connections)
    async with aiohttp.ClientSession(connector=connector) as session:
        for pdb_id in pdb_ids:
            task = asyncio.create_task(
                async_download_pdbx(
                    pdb_id=pdb_id,
                    download_dir=download_dir,
                    session=session,
                    sem=sem,
                ))
            tasks.append(task)

        responses = [
            await f
            for f in tqdm.tqdm(asyncio.as_completed(tasks), total=len(tasks))
        ]

        # try:
        #     await asyncio.gather(*tasks)
        # except Exception as e:
        #     print(repr(e))

        # No progress bar.
        # await asyncio.gather(*tasks)


def download_x_ray_pdbxs(download_dir: str,
                         maximum_resolution: float = 1.5) -> None:

    print("Querying PDB IDs of maximum resolution {}...".format(
        maximum_resolution))
    pdb_ids = query_x_ray_entries(maximum_resolution=maximum_resolution)

    print("Downloading {} PDBxs...".format(len(pdb_ids)))
    time_start = time.time()
    # Python 3.7+
    # asyncio.run(async_download_pdbs(pdb_ids=pdb_ids, download_dir=download_dir))
    # Python 3.5-3.6
    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        async_download_pdbxs(pdb_ids=pdb_ids, download_dir=download_dir))
    time_end = time.time()
    time_elapsed = time_end - time_start
    print("Download Time Elapsed: {:02d}:{:02d}:{:02d}".format(
        int(time_elapsed // 3600), int(time_elapsed % 3600 // 60),
        int(time_elapsed % 60 // 1)))



from typing import List, Tuple, Optional
import os
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from matplotlib import cm
import matplotlib.colors as mplcolors
from ramachandran.io import read_residue_torsion_collection_from_file


def get_coordinates_on_reference_map(
        phi_psi_angle: Tuple[float, float],
        reference_map: np.ndarray) -> Tuple[int, int]:

    phi = phi_psi_angle[0]
    psi = phi_psi_angle[1]

    height = reference_map.shape[0]
    width = reference_map.shape[1]

    i = int((180 - psi) / 360 * height)
    j = int((phi + 180) / 360 * width)

    # If i or j == resolution, adjust it.
    if i == height:
        i = height - 1
    if j == width:
        j = width - 1

    return (i, j)


def create_ramachandran_plot(phi_psi_angles: List[Tuple[float, float]],
                             plot_file_path: str,
                             reference_map: Optional[np.ndarray] = None,
                             cmap: Optional[mplcolors.ListedColormap] = None,
                             protein_name: Optional[str] = None,
                             rendering_interpolation: bool = True) -> None:

    phi_psi_angles_numpy = np.array(phi_psi_angles)
    x_numpy = phi_psi_angles_numpy[:, 0]
    y_numpy = phi_psi_angles_numpy[:, 1]

    fig = plt.figure(figsize=(10, 10))

    ax = fig.add_subplot(111)
    if protein_name is not None:
        ax.set_title(protein_name, fontsize=24)

    interpolation = None
    if rendering_interpolation is True:
        interpolation = "bilinear"

    if reference_map is not None:

        percentile_1 = np.percentile(reference_map, 60)
        percentile_2 = np.percentile(reference_map, 90)

        ax.imshow(np.rot90(reference_map),
                  interpolation=interpolation,
                  cmap=cmap,
                  norm=mplcolors.BoundaryNorm(
                      boundaries=[0, percentile_1, percentile_2, 1],
                      ncolors=cmap.N),
                  origin="upper",
                  extent=(-180, 180, -180, 180))

        # Find outliers
        outliers_idx = []
        for i, phi_psi_angle in enumerate(phi_psi_angles):

            map_i, map_j = get_coordinates_on_reference_map(
                phi_psi_angle=phi_psi_angle,
                reference_map=np.rot90(reference_map))
            if np.rot90(reference_map)[map_i, map_j] < percentile_1:
                outliers_idx.append(i)

        x_outliers_numpy = x_numpy[outliers_idx]
        y_outliers_numpy = y_numpy[outliers_idx]

        x_numpy = np.delete(x_numpy, outliers_idx)
        y_numpy = np.delete(y_numpy, outliers_idx)

        ax.scatter(x_outliers_numpy,
                   y_outliers_numpy,
                   s=20,
                   color="red",
                   edgecolors="black")

    ax.scatter(x_numpy, y_numpy, s=20, color="blue", edgecolors="black")

    ax.set_xlim(-180, 180)
    ax.set_ylim(-180, 180)

    ax.xaxis.set_major_locator(ticker.MultipleLocator(45))
    ax.yaxis.set_major_locator(ticker.MultipleLocator(45))

    ax.xaxis.set_tick_params(labelsize=12)
    ax.yaxis.set_tick_params(labelsize=12)

    ax.plot([-180, 180], [0, 0], "--", linewidth=0.5, color="black")
    ax.plot([0, 0], [-180, 180], "--", linewidth=0.5, color="black")

    ax.set_xlabel(r"${\phi}$", fontsize=18, fontweight="bold")
    ax.set_ylabel(r"${\psi}$", fontsize=18, fontweight="bold")

    fig.savefig(plot_file_path, format="svg", dpi=600, bbox_inches="tight")
    plt.close()

    return


def create_ramachandran_plots_from_file(
        file_path: str,
        save_dir_path: str,
        # reference_map_type: Optional[str] = "unsmoothed",
        protein_name: Optional[str] = None,
        rendering_interpolation: bool = False) -> None:

    if not os.path.exists(save_dir_path):
        os.makedirs(save_dir_path)

    residue_torsion_collection = read_residue_torsion_collection_from_file(
        file_path=file_path)

    phi_psi_angles_general = residue_torsion_collection.collect_torsion_angles_general(
    )
    phi_psi_angles_gly = residue_torsion_collection.collect_torsion_angles_gly(
    )
    phi_psi_angles_pro = residue_torsion_collection.collect_torsion_angles_pro(
    )
    phi_psi_angles_prepro = residue_torsion_collection.collect_torsion_angles_prepro(
    )

    phi_psi_angles_list = [
        phi_psi_angles_general, phi_psi_angles_gly, phi_psi_angles_pro,
        phi_psi_angles_prepro
    ]

    package_dir, filename = os.path.split(__file__)

    # Using unsmoothed probability.npz is problematic because
    # many probabilities are exactly zeros and thus the many percentiles are exactly zeros.
    # Plotting these zero values is very problematic.
    # Gaussian density is fine because none of the probability density values are exactly zero.

    # if reference_map_type == "unsmoothed":
    #     npz_file_path = os.path.join(package_dir, "data", "probability.npz")
    #     npz_file = np.load(npz_file_path)
    # elif reference_map_type == "smoothed":
    #     npz_file_path = os.path.join(package_dir, "data", "gaussian_density.npz")
    #     npz_file = np.load(npz_file_path)
    # else:
    #     raise RuntimeError("Unsupported reference map type.")

    npz_file_path = os.path.join(package_dir, "data", "gaussian_density.npz")
    npz_file = np.load(npz_file_path)

    reference_map_general = npz_file["general"]
    reference_map_gly = npz_file["gly"]
    reference_map_pro = npz_file["pro"]
    reference_map_prepro = npz_file["prepro"]

    reference_map_list = [
        reference_map_general, reference_map_gly, reference_map_pro,
        reference_map_prepro
    ]

    # Using Erdős Gábor's cmaps.
    # https://github.com/gerdos/PyRAMA/blob/301df17e5f2c32544b34321c4f8b0254697183ce/pyrama/config.py
    cmap_general = mplcolors.ListedColormap(['#FFFFFF', '#B3E8FF', '#7FD9FF'])
    cmap_gly = mplcolors.ListedColormap(['#FFFFFF', '#FFE8C5', '#FFCC7F'])
    cmap_pro = mplcolors.ListedColormap(['#FFFFFF', '#D0FFC5', '#7FFF8C'])
    cmap_prepro = mplcolors.ListedColormap(['#FFFFFF', '#B3E8FF', '#7FD9FF'])

    cmap_list = [cmap_general, cmap_gly, cmap_pro, cmap_prepro]

    filename_list = ["general.svg", "gly.svg", "pro.svg", "prepro.svg"]
    file_path_list = [
        os.path.join(save_dir_path, filename) for filename in filename_list
    ]

    for phi_psi_angles, reference_map, cmap, file_path in zip(
            phi_psi_angles_list, reference_map_list, cmap_list,
            file_path_list):

        create_ramachandran_plot(
            phi_psi_angles=phi_psi_angles,
            reference_map=reference_map,
            cmap=cmap,
            plot_file_path=file_path,
            rendering_interpolation=rendering_interpolation,
            protein_name=protein_name)

    return



from typing import List, Tuple, Optional
import os
import math
from tqdm import tqdm
from multiprocessing import Pool
import numpy as np
from scipy import stats
from ramachandran.io import read_residue_torsion_collection_from_directory


def count_torsion_angles(phi_psi_angles: List[Tuple[float, float]],
                         resolution: int = 90,
                         normalized: bool = False) -> np.ndarray:
    """Count the torsion angles for histogram. No smoothing function was applied.

    Args:
        phi_psi_angles (List[Tuple[float, float]]): Torsion phi psi angles.
        resolution (int, optional): Resolution of the counts. Defaults to 90.
        normalized (bool, optional): Normalize the counts to density. Defaults to False.

    Returns:
        np.ndarray: A Numpy array of shape [resolution, resolution].
    """

    counts = np.zeros(shape=(resolution, resolution), dtype=np.uint32)

    for phi, psi in phi_psi_angles:

        i = int((phi + 180) / 360 * resolution)
        j = int((psi + 180) / 360 * resolution)
        # i = int((180 - psi) / 360 * resolution)
        # j = int((phi + 180) / 360 * resolution)

        # If i or j == resolution, adjust it.
        if i == resolution:
            i = resolution - 1
        if j == resolution:
            j = resolution - 1

        counts[i, j] += 1

    if normalized == True:
        counts = counts.astype(np.float32) / np.sum(counts)

    return counts


def compute_gaussian_kde_density(phi_psi_angles: List[Tuple[float, float]],
                                 resolution: int = 180) -> np.ndarray:
    """Use Gaussian kernel density smoothed estimation for the torsion angles.

    Args:
        phi_psi_angles (List[Tuple[float, float]]): Torsion phi psi angles.
        resolution (int, optional): Resolution of the density. Defaults to 180.

    Returns:
        np.ndarray: A Numpy array of shape [resolution, resolution].
    """

    # Turn the list of tuples to Numpy array
    # Shape: [2, len(phi_psi_angles)]
    values = np.array(phi_psi_angles).T

    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html
    xmin = -180
    xmax = 180
    ymin = -180
    ymax = 180

    X, Y = np.mgrid[xmin:xmax:(1j * resolution), ymin:ymax:(1j * resolution)]
    positions = np.vstack([X.ravel(), Y.ravel()])

    kernel = stats.gaussian_kde(values)
    Z = np.reshape(kernel(positions).T, X.shape)

    return Z


def count_torsion_angles_from_directory(
    dir_path: str,
    save_file_path: Optional[str] = None,
    b_factor_threshold: float = 30,
    resolution: int = 180,
    num_processes: int = 4
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:

    residue_torsion_collection = read_residue_torsion_collection_from_directory(
        dir_path=dir_path,
        b_factor_threshold=b_factor_threshold,
        num_processes=num_processes)

    # Counting residue torsions
    # GLY
    phi_psi_angles = residue_torsion_collection.collect_torsion_angles_gly()
    print("Couting torsions for GLY, {} samples, this might take a while...".
          format(len(phi_psi_angles)))
    counts_gly = count_torsion_angles(phi_psi_angles=phi_psi_angles,
                                      resolution=resolution,
                                      normalized=True)
    # PRO
    phi_psi_angles = residue_torsion_collection.collect_torsion_angles_pro()
    print("Couting torsions for PRO, {} samples, this might take a while...".
          format(len(phi_psi_angles)))
    counts_pro = count_torsion_angles(phi_psi_angles=phi_psi_angles,
                                      resolution=resolution,
                                      normalized=True)
    # pre-PRO
    phi_psi_angles = residue_torsion_collection.collect_torsion_angles_prepro()
    print(
        "Couting torsions for pre-PRO, {} samples, this might take a while...".
        format(len(phi_psi_angles)))
    counts_prepro = count_torsion_angles(phi_psi_angles=phi_psi_angles,
                                         resolution=resolution,
                                         normalized=True)
    # General
    phi_psi_angles = residue_torsion_collection.collect_torsion_angles_general(
    )
    print(
        "Couting torsions for General, {} samples, this might take a while...".
        format(len(phi_psi_angles)))
    counts_general = count_torsion_angles(phi_psi_angles=phi_psi_angles,
                                          resolution=resolution,
                                          normalized=True)

    if save_file_path is not None:

        dir_path = os.path.dirname(save_file_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        np.savez(save_file_path,
                 gly=counts_gly,
                 pro=counts_pro,
                 prepro=counts_prepro,
                 general=counts_general)

    return (counts_gly, counts_pro, counts_prepro, counts_general)


def compute_gaussian_kde_densities_from_directory(
    dir_path: str,
    save_file_path: Optional[str] = None,
    b_factor_threshold: float = 30,
    resolution: int = 180,
    num_processes: int = 4
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:

    residue_torsion_collection = read_residue_torsion_collection_from_directory(
        dir_path=dir_path,
        b_factor_threshold=b_factor_threshold,
        num_processes=num_processes)

    # Computing Gaussian KDE
    # This is a single-thread implementation and is somewhat slow.
    # GLY
    phi_psi_angles = residue_torsion_collection.collect_torsion_angles_gly()
    print(
        "Computing Gaussian kernel for GLY, {} samples, this might take a while..."
        .format(len(phi_psi_angles)))
    gaussian_density_gly = compute_gaussian_kde_density(
        phi_psi_angles=phi_psi_angles, resolution=resolution)
    # PRO
    phi_psi_angles = residue_torsion_collection.collect_torsion_angles_pro()
    print(
        "Computing Gaussian kernel for PRO, {} samples, this might take a while..."
        .format(len(phi_psi_angles)))
    gaussian_density_pro = compute_gaussian_kde_density(
        phi_psi_angles=phi_psi_angles, resolution=resolution)
    # pre-PRO
    phi_psi_angles = residue_torsion_collection.collect_torsion_angles_prepro()
    print(
        "Computing Gaussian kernel for pre-PRO, {} samples, this might take a while..."
        .format(len(phi_psi_angles)))
    gaussian_density_prepro = compute_gaussian_kde_density(
        phi_psi_angles=phi_psi_angles, resolution=resolution)
    # General
    phi_psi_angles = residue_torsion_collection.collect_torsion_angles_general(
    )
    print(
        "Computing Gaussian kernel for general, {} samples, this might take a while..."
        .format(len(phi_psi_angles)))
    gaussian_density_general = compute_gaussian_kde_density(
        phi_psi_angles=phi_psi_angles, resolution=resolution)

    # print(np.sum(gaussian_density_gly))
    # print(np.sum(gaussian_density_pro))
    # print(np.sum(gaussian_density_prepro))
    # print(np.sum(gaussian_density_general))

    if save_file_path is not None:

        dir_path = os.path.dirname(save_file_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        np.savez(save_file_path,
                 gly=gaussian_density_gly,
                 pro=gaussian_density_pro,
                 prepro=gaussian_density_prepro,
                 general=gaussian_density_general)

    return (gaussian_density_gly, gaussian_density_pro,
            gaussian_density_prepro, gaussian_density_general)



from __future__ import annotations
from typing import List, Optional, Sequence, Set, Tuple

RESIDUE_TYPES = {
    "ALA", "CYS", "ASP", "GLU", "PHE", "GLY", "HIS", "ILE", "LYS", "LEU",
    "MET", "ASN", "PRO", "GLN", "ARG", "SER", "THR", "VAL", "TRP", "TYR"
}


class ResidueTorsion(object):
    """Data structure for storing residue torsion information.
    """
    def __init__(self,
                 phi: float,
                 psi: float,
                 omega: float = 180,
                 residue_type: Optional[str] = None,
                 is_pre_proline: Optional[bool] = None) -> None:

        self.phi = phi
        self.psi = psi
        self.omega = omega
        self.residue_type = residue_type
        self.is_pre_proline = is_pre_proline

    def _sanity_check(self):

        if self.phi < -180 or self.phi > 180:
            raise ValueError(
                f"Torsion angle phi is defined in a range of [-180, 180]. Got {self.phi}"
            )

        if self.psi < -180 or self.psi > 180:
            raise ValueError(
                f"Torsion angle psi is defined in a range of [-180, 180]. Got {self.psi}"
            )

        if self.omega < -180 or self.omega > 180:
            raise ValueError(
                f"Torsion angle omega is defined in a range of [-180, 180]. Got {self.omega}"
            )

        if self.residue_type is not None and self.residue_type not in RESIDUE_TYPES:
            raise ValueError(
                f"Unsupported amino acid type. Got {self.residue_type}")


class ResidueTorsionCollection(object):
    """A collection of ResidueTorsion objects for analysis.
    """
    def __init__(
            self,
            residue_torsions: Optional[List[ResidueTorsion]] = None) -> None:

        # The ResidueTorsionCollection object will own a copy of the data.
        if residue_torsions is None:
            self.residue_torsions = []
        else:
            self.residue_torsions = residue_torsions.copy()

    def append(self, residue_torsion: ResidueTorsion) -> None:

        self.residue_torsions.append(residue_torsion)

    def collect_torsion_angles_by_residue_types(
            self,
            residue_types: Set[str],
            is_pre_proline: Optional[bool] = None
    ) -> List[Tuple[float, float]]:
        """Collect torsion angles phi and psi by residue type.

        Args:
            residue_type (Set[str]): A sequence of residue types that needs to be collected.

        Returns:
            List[Tuple[float, float]]: A list of phi psi angle tuples.
        """

        phi_psi_angles = []

        for residue_torsion in self.residue_torsions:
            if residue_torsion.residue_type in residue_types:
                if is_pre_proline is not None and residue_torsion.is_pre_proline != is_pre_proline:
                    continue
                phi_psi_angles.append(
                    (residue_torsion.phi, residue_torsion.psi))

        return phi_psi_angles

    def collect_torsion_angles_pro(self) -> List[Tuple[float, float]]:

        return self.collect_torsion_angles_by_residue_types(
            residue_types={"PRO"}, is_pre_proline=False)

    def collect_torsion_angles_prepro(self) -> List[Tuple[float, float]]:

        return self.collect_torsion_angles_by_residue_types(
            residue_types=RESIDUE_TYPES, is_pre_proline=True)

    def collect_torsion_angles_gly(self) -> List[Tuple[float, float]]:

        return self.collect_torsion_angles_by_residue_types(
            residue_types={"GLY"}, is_pre_proline=False)

    def collect_torsion_angles_general(self) -> List[Tuple[float, float]]:

        general_residue_types = RESIDUE_TYPES.copy()
        general_residue_types.remove("PRO")
        general_residue_types.remove("GLY")

        return self.collect_torsion_angles_by_residue_types(
            residue_types=general_residue_types, is_pre_proline=False)

    # Override +=
    def __iadd__(self,
                 other: ResidueTorsionCollection) -> ResidueTorsionCollection:

        self.residue_torsions.extend(other.residue_torsions)

        return self

    # Override +
    def __add__(self,
                other: ResidueTorsionCollection) -> ResidueTorsionCollection:

        return ResidueTorsionCollection(
            residue_torsions=self.residue_torsions + other.residue_torsions)



from typing import Any, Dict, Optional, List
import os
import numpy as np
from ramachandran.geometry import protein_backbone_dihedral_angle_phi, protein_backbone_dihedral_angle_psi
from ramachandran.torsion import ResidueTorsionCollection, ResidueTorsion
from tqdm import tqdm
from multiprocessing import Pool
from functools import partial


def read_pdb(pdb_file_path: str) -> Dict[Any, Any]:
    """Read the basic information of a protein from a PDB file. 
        The function does minimal format checking for the PDB format.
        Please make sure the PDB format is correct before calling the function.

    Args:
        pdb_file_path (str): file path to the PDB file.

    Returns:
        Dict[Any, Any]: A dictionary containing the basic protein information.
    """
    sequence = {}

    with open(pdb_file_path, "r") as fp:

        for line in fp:
            # Check if the line is SEQRES
            if line[0:6] == "SEQRES":
                items = line.split()

                chain_identifier = items[2]
                amino_acids = items[4:]

                if chain_identifier not in sequence:
                    sequence[chain_identifier] = []
                sequence[chain_identifier] += amino_acids

    protein = {}

    with open(pdb_file_path, "r") as fp:

        for line in fp:
            # Check if the line is ATOM
            # 1-4 Record Type
            if line[0:4] == "ATOM":
                # It is not best practice for dealing with PDB format.
                # Use it for the ATOM section for now.
                items = line.split()

                try:
                    atom_serial_number = int(items[1])
                except:
                    raise RuntimeError(
                        "Atom serial number has to be integer. Got {}".format(
                            items[1]))

                atom_name = items[2]
                residue_name = items[3]
                chain_identifier = items[4]
                try:
                    residue_sequence_number = int(items[5])
                except:
                    raise RuntimeError(
                        "Residue sequence number has to be integer.")

                try:
                    x = float(items[6])
                    y = float(items[7])
                    z = float(items[8])
                except:
                    raise RuntimeError(
                        "Atom coordinate has to be a real value.")

                try:
                    b_factor = float(items[10])
                except:
                    raise RuntimeError("B factor has to be a real value.")

                atom_coodinates = ((x, y, z), b_factor)

                if chain_identifier not in protein:
                    protein[chain_identifier] = {}

                if residue_sequence_number not in protein[chain_identifier]:
                    protein[chain_identifier][residue_sequence_number] = {}

                if "residue" not in protein[chain_identifier][
                        residue_sequence_number]:
                    protein[chain_identifier][residue_sequence_number][
                        "residue"] = residue_name

                if "is_pre_proline" not in protein[chain_identifier][
                        residue_sequence_number]:

                    chain_sequence = sequence[chain_identifier]
                    num_residues = len(chain_sequence)

                    if residue_sequence_number == num_residues:
                        protein[chain_identifier][residue_sequence_number][
                            "is_pre_proline"] = False
                    else:
                        protein[chain_identifier][residue_sequence_number][
                            "is_pre_proline"] = (
                                chain_sequence[residue_sequence_number] ==
                                "PRO")

                protein[chain_identifier][residue_sequence_number][
                    atom_name] = atom_coodinates

    return protein


def read_pdbx(pdbx_file_path: str) -> Dict[Any, Any]:
    """Read the basic information of a protein from a PDBx file. 
        The function does minimal format checking for the PDBx format.
        Please make sure the PDBx format is correct before calling the function.

    Args:
        pdbx_file_path (str): file path to the PDBx file.

    Returns:
        Dict[Any, Any]: A dictionary containing the basic protein information.
    """

    sequence = {}

    with open(pdbx_file_path, "r") as fp:

        is_sequence = False

        for line in fp:

            if line.strip() == "#":
                is_sequence = False

            # Check if the line is SEQRES
            if is_sequence:
                items = line.split()
                assert len(items) == 4, line

                chain_identifier = items[0]
                amino_acid = items[2]

                if chain_identifier not in sequence:
                    sequence[chain_identifier] = []
                sequence[chain_identifier].append(amino_acid)

            if line.strip() == "_entity_poly_seq.hetero":
                is_sequence = True

    protein = {}

    # http://ww1.iucr.org/iucr-top/cif/mmcif/workshop/mmCIF-tutorials/intro/atom.htm
    with open(pdbx_file_path, "r") as fp:

        is_atom = False

        for line in fp:

            if line.strip() == "#":
                is_atom = False

            # Check if the line is ATOM
            # 1-4 Record Type
            if is_atom and line[0:4] == "ATOM":
                # It is not best practice for dealing with PDB format.
                # Use it for the ATOM section for now.
                items = line.split()

                try:
                    atom_serial_number = int(items[1])
                except:
                    raise RuntimeError(
                        "Atom serial number has to be integer. Got {}".format(
                            pdbx_file_path))

                atom_name = items[3]
                residue_name = items[5]
                chain_identifier = items[6]
                chain_sequence_identifier = items[7]

                try:
                    residue_sequence_number = int(items[8])
                except:
                    raise RuntimeError(
                        "Residue sequence number has to be integer.")

                try:
                    x = float(items[10])
                    y = float(items[11])
                    z = float(items[12])
                except:
                    raise RuntimeError(
                        "Atom coordinate has to be a real value.")

                try:
                    b_factor = float(items[14])
                except:
                    raise RuntimeError("B factor has to be a real value.")

                atom_coodinates = ((x, y, z), b_factor)

                if chain_identifier not in protein:
                    protein[chain_identifier] = {}

                if residue_sequence_number not in protein[chain_identifier]:
                    protein[chain_identifier][residue_sequence_number] = {}

                if "residue" not in protein[chain_identifier][
                        residue_sequence_number]:
                    protein[chain_identifier][residue_sequence_number][
                        "residue"] = residue_name

                if "is_pre_proline" not in protein[chain_identifier][
                        residue_sequence_number]:

                    chain_sequence = sequence[chain_sequence_identifier]
                    num_residues = len(chain_sequence)

                    if residue_sequence_number == num_residues:
                        protein[chain_identifier][residue_sequence_number][
                            "is_pre_proline"] = False
                    else:
                        protein[chain_identifier][residue_sequence_number][
                            "is_pre_proline"] = (
                                chain_sequence[residue_sequence_number] ==
                                "PRO")

                protein[chain_identifier][residue_sequence_number][
                    atom_name] = atom_coodinates

            if line.strip() == "_atom_site.pdbx_PDB_model_num":
                is_atom = True

    return protein


def read_residue_torsion_collection_from_protein(
        protein: Dict[Any, Any],
        b_factor_threshold: Optional[float] = None
) -> ResidueTorsionCollection:
    """Read torsion collection from a protein dictionary.

    Args:
        protein (Dict[Any, Any]): A protein dictionary.
        b_factor_threshold (Optional[float], optional): B-factor indicating the uncertainty of the atom coordinates. Defaults to None.

    Returns:
        ResidueTorsionCollection: A residue torsion collection for the protein given.
    """

    residue_torsion_collection = ResidueTorsionCollection()

    for chain_identifier in protein:

        chain = protein[chain_identifier]
        for residue_sequence_number in chain:
            residue = chain[residue_sequence_number]
            residue_name = residue["residue"]
            # Skip the first, the last, and problematic residues
            if residue_sequence_number - 1 not in chain or residue_sequence_number + 1 not in chain:
                continue
            last_residue = chain[residue_sequence_number - 1]
            next_residue = chain[residue_sequence_number + 1]
            next_residue_name = next_residue["residue"]

            # Skip the residues that has missing information to compute dihedral angles
            if "N" not in residue or "CA" not in residue or "C" not in residue or "C" not in last_residue or "N" not in next_residue:
                continue

            n, b_factor_n = residue["N"]
            c_alpha, b_factor_c_alpha = residue["CA"]
            c, b_factor_c = residue["C"]
            c_minus, b_factor_c_minus = last_residue["C"]
            n_plus, b_factor_n_plus = next_residue["N"]

            if b_factor_threshold is not None:
                if b_factor_n > b_factor_threshold or b_factor_c_alpha > b_factor_threshold or b_factor_c > b_factor_threshold or b_factor_c_minus > b_factor_threshold or b_factor_n_plus > b_factor_threshold:
                    continue

            phi = protein_backbone_dihedral_angle_phi(c_minus=c_minus,
                                                      n=n,
                                                      c_alpha=c_alpha,
                                                      c=c)
            psi = protein_backbone_dihedral_angle_psi(n=n,
                                                      c_alpha=c_alpha,
                                                      c=c,
                                                      n_plus=n_plus)

            phi = np.rad2deg(phi)
            psi = np.rad2deg(psi)
            is_pre_proline = (next_residue_name == "PRO")

            residue_torsion = ResidueTorsion(phi=phi,
                                             psi=psi,
                                             residue_type=residue_name,
                                             is_pre_proline=is_pre_proline)

            residue_torsion_collection.append(residue_torsion=residue_torsion)

    return residue_torsion_collection


def read_residue_torsion_collection_from_file(
        file_path: str,
        b_factor_threshold: Optional[float] = None
) -> ResidueTorsionCollection:
    """Read torsion collection from a PDB or PDBx file.

    Args:
        file_path (str): file path to a PDB or PDBx file.
        b_factor_threshold (Optional[float], optional): B-factor indicating the uncertainty of the atom coordinates. Defaults to None.

    Raises:
        RuntimeError: When non-PDB or non-PDBx files were provided, raise RuntimeError.

    Returns:
        ResidueTorsionCollection: A residue torsion collection for the PDB or PDBx file given.
    """

    _, file_extension = os.path.splitext(file_path)

    if file_extension == ".pdb":
        protein = read_pdb(pdb_file_path=file_path)
    elif file_extension == ".cif":
        protein = read_pdbx(pdbx_file_path=file_path)
    else:
        raise RuntimeError(
            "Only files with extensions of pdb and cif are supported.")

    residue_torsion_collection = read_residue_torsion_collection_from_protein(
        protein=protein, b_factor_threshold=b_factor_threshold)

    return residue_torsion_collection


def read_residue_torsion_collection_from_file_resolved(
        file_path: str,
        b_factor_threshold: Optional[float] = None
) -> ResidueTorsionCollection:
    try:
        residue_torsion_collection = read_residue_torsion_collection_from_file(
            file_path=file_path, b_factor_threshold=b_factor_threshold)
    except:
        residue_torsion_collection = ResidueTorsionCollection()
    return residue_torsion_collection


def read_residue_torsion_collection_from_files(
        file_paths: List[str],
        num_processes: int = 4,
        b_factor_threshold: Optional[float] = None
) -> ResidueTorsionCollection:

    pool = Pool(processes=num_processes)

    residue_torsion_collection_all = ResidueTorsionCollection()

    partial_func = partial(read_residue_torsion_collection_from_file_resolved,
                           b_factor_threshold=b_factor_threshold)

    jobs = [
        pool.apply_async(func=partial_func, args=(file_path, ))
        for file_path in file_paths
    ]

    pool.close()

    for job in tqdm(jobs):
        residue_torsion_collection_all += job.get()

    return residue_torsion_collection_all


def read_residue_torsion_collection_from_directory(
        dir_path: str,
        b_factor_threshold: float = 30,
        num_processes: int = 4) -> ResidueTorsionCollection:

    valid_file_paths = []

    for f in os.listdir(dir_path):

        _, file_extension = os.path.splitext(f)
        file_path = os.path.join(dir_path, f)

        if os.path.isfile(file_path) and (file_extension == ".pdb"
                                          or file_extension == ".cif"):
            valid_file_paths.append(file_path)

    residue_torsion_collection = read_residue_torsion_collection_from_files(
        file_paths=valid_file_paths,
        num_processes=num_processes,
        b_factor_threshold=b_factor_threshold)

    return residue_torsion_collection



from collections.abc import Sequence
from typing import List, Tuple, Optional
import os
import numpy as np
from numpy import linalg as LA


def polymer_physics_dihedral_angle(a: Sequence, b: Sequence,
                                   c: Sequence) -> float:
    """Given a plane defined by vectors a and b, and a plane defined by vectors b and c, compute the dihedral angle of the two planes. The dihedral angle is in a range of [-pi, pi].

    Args:
        a (Sequence): [description]
        b (Sequence): [description]
        c (Sequence): [description]

    Returns:
        float: [description]
    """

    # https://en.wikipedia.org/wiki/Dihedral_angle#In_polymer_physics

    axb = np.cross(a=a, b=b)
    norm_axb = LA.norm(axb)
    bxc = np.cross(a=b, b=c)
    norm_bxc = LA.norm(bxc)

    norm_b = LA.norm(b)

    cos_phi = np.inner(axb, bxc) / (norm_axb * norm_bxc)

    # We use the definition of cross product to compute sin phi.
    # axb, b, bxc is a right handed system.
    sin_phi = np.inner(b, np.cross(a=axb,
                                   b=bxc)) / (norm_b * norm_axb * norm_bxc)

    angle = np.arctan2(sin_phi, cos_phi)

    return angle


def protein_backbone_dihedral_angle_phi(c_minus: Sequence, n: Sequence,
                                        c_alpha: Sequence,
                                        c: Sequence) -> float:
    """Given the atom coordinates of C_(-1), N, C_alpha, and C, compute the dihedral angle phi.

    Args:
        c_minus (Sequence): [description]
        n (Sequence): [description]
        c_alpha (Sequence): [description]
        c (Sequence): [description]

    Returns:
        float: [description]
    """

    # Compute three vectors
    b1 = np.array(n) - np.array(c_minus)
    b2 = np.array(c_alpha) - np.array(n)
    b3 = np.array(c) - np.array(c_alpha)

    angle = polymer_physics_dihedral_angle(a=b1, b=b2, c=b3)

    return angle


def protein_backbone_dihedral_angle_psi(n: Sequence, c_alpha: Sequence,
                                        c: Sequence, n_plus) -> float:
    """Given the atom coordinates of N, C_alpha, C, and N_(+1) compute the dihedral angle psi.

    Args:
        n (Sequence): [description]
        c_alpha (Sequence): [description]
        c (Sequence): [description]
        n_plus ([type]): [description]

    Returns:
        float: [description]
    """

    # Compute three vectors
    b1 = np.array(c_alpha) - np.array(n)
    b2 = np.array(c) - np.array(c_alpha)
    b3 = np.array(n_plus) - np.array(c)

    angle = polymer_physics_dihedral_angle(a=b1, b=b2, c=b3)

    return angle



import setuptools

with open("README.md", "r") as fhand:
  long_description = fhand.read()

setuptools.setup(
    name="ramachandran",
    version="0.0.1",
    author="Lei Mao",
    author_email="dukeleimao@gmail.com",
    description="Generating Ramachandran plot and statistics.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/leimao/Ramachandran",
    # packages=setuptools.find_packages(),
    packages=["ramachandran"],
    package_dir={"ramachandran": "ramachandran"},
    package_data={"ramachandran": ["data/*.npz"]},
    scripts=["scripts/ramachandran_plot"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)
